//==============================================================================
//	
//	Copyright (c) 2002-
//	Authors:
//	* Dave Parker <david.parker@comlab.ox.ac.uk> (University of Oxford, formerly University of Birmingham)
//	
//------------------------------------------------------------------------------
//	
//	This file is part of PRISM.
//	
//	PRISM is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	PRISM is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with PRISM; if not, write to the Free Software Foundation,
//	Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//==============================================================================

package jdd;

import java.util.Collections;
import java.util.Comparator;
import java.util.Vector;

import prism.PrismException;

public class JDDVars
{
	private Vector<JDDNode> vars;
	
	private native long DDV_BuildArray();
	private native void DDV_FreeArray(long a);
	private native int DDV_GetIndex(long dd);
	private native int DDV_GetLevel(long dd);

	static
	{
		try {
			System.loadLibrary("jdd");
		}
		catch (UnsatisfiedLinkError e) {
			System.out.println(e);
			System.exit(1);
		}
	}

	public class ArrayRef {
		private long array;

		public long array() {
			return array;
		}
		
		public void clear() {
			DDV_FreeArray(array);
			array = 0;
		}
	}
	
	public JDDVars()
	{
		vars = new Vector();
	}
		
	public void addVar(JDDNode var)
	{
		vars.addElement(var);
	}
	
	public void addVars(JDDVars ddv)
	{
		int i;
		
		vars.addAll(ddv.vars);
	}
	
	public void removeVars(JDDVars ddv)
	{
		int i;
		
		vars.removeAll(ddv.vars);
	}
	
	public int getNumVars()
	{
		return vars.size();
	}

	public JDDNode getVar(int i)
	{
		return (JDDNode)vars.elementAt(i);
	}
	
	public long getVarPtr(int i)
	{
		return ((JDDNode)vars.elementAt(i)).ptr();
	}
	
	public int getVarIndex(int i)
	{
		return DDV_GetIndex(((JDDNode)vars.elementAt(i)).ptr());
	}

	public int getVarLevel(int i)
	{
		return DDV_GetLevel(((JDDNode)vars.elementAt(i)).ptr());
	}
	
	public int getMinVarLevel()
	{
		int i, j, n, min;
		n = vars.size();
		if (n == 0) return -1;
		min = getVarLevel(0);
		for (i = 1; i < n; i++) {
			j = getVarLevel(i);
			if (j < min) min = j;
		}
		return min;
	}
	
	public int getMaxVarLevel()
	{
		int i, j, n, max;
		n = vars.size();
		if (n == 0) return -1;
		max = getVarLevel(0);
		for (i = 1; i < n; i++) {
			j = getVarLevel(i);
			if (j > max) max = j;
		}
		return max;
	}
	
	public void refAll()
	{
		int i;
		
		for (i = 0; i < vars.size(); i++) {
			JDD.Ref((JDDNode)vars.elementAt(i));
		}
	}
	
	public void derefAll()
	{
		int i;
		
		for (i = 0; i < vars.size(); i++) {
			JDD.Deref((JDDNode)vars.elementAt(i));
		}
	}
	
	public ArrayRef array()
	{
		ArrayRef ref = new ArrayRef();
		ref.array = DDV_BuildArray();
		return ref;
	}
	
	public ArrayRef sortedArray()
	{
		JDDVars tmp = new JDDVars();
		tmp.addVars(this);
		tmp.refAll();
		tmp.sort();
		ArrayRef ref = tmp.array();
		tmp.clear();
		
		return ref;
	}
	
	public JDDVars toSorted() {
		JDDVars result = new JDDVars();
		result.addVars(this);
		result.refAll();
		return result;
	}
	
	public void sort() 
	{
		Collections.sort(vars, new Comparator<JDDNode>() {
			@Override
			public int compare(JDDNode a, JDDNode b)
			{
				return new Integer(a.getLevel()).compareTo(b.getLevel());
			}
		});	
	}
	
	public boolean isSorted() 
	{
		for (int i=1;i<vars.size();i++)
		{
			if (vars.get(i-1).getLevel() > vars.get(i).getLevel()) {
				return false;
			}
		}
		return true;
	}
	
	public void checkIsSorted() throws PrismException 
	{
		if (!isSorted())
			throw new PrismException("JDDVars is not sorted!");
	}
	
	public int n()
	{
		return vars.size();
	}

	/** Clears all variables (incl. deref) */
	public void clear()
	{
		derefAll();
		vars.clear();
	}

	public String toString()
	{
		int i;
		String s = "{";
		
		for (i = 0; i < vars.size(); i++) {
			if (i>0) s = s + ", ";
			int index = getVarIndex(i);
			int level = getVarLevel(i);
			
			if (index == level)
				s = s + index;
			else 
				s = s + index + "@" + level;
		}
		s += "}";
		
		return s;
	}
}

//------------------------------------------------------------------------------
